@@ -9,13 +9,15 @@
 //! See the structs section below to find out more about each struct
 
 use crate::event::{Error, Event, Result, Status};
-use crate::regex;
 use crate::row::Row;
-use crate::utils::{width_char, Loc, Size};
+use crate::utils::{filetype, width_char, Loc, Size};
+use crate::{regex, st};
+use std::collections::HashMap;
 use std::fs;
+use std::path::Path;
 
 /// A struct that stores information about a file
-#[derive(Debug, PartialEq, Clone)]
+#[derive(Debug, Clone, PartialEq)]
 pub struct FileInfo {
     /// The file name of the document
     pub file: Option<String>,
@@ -55,6 +57,12 @@ pub struct Document {
     pub cursor: Loc,
     /// Stores information about scrolling
     pub offset: Loc,
+    /// Render cache space for optimisation purposes
+    #[cfg(feature = "syntax_highlighting")]
+    pub render: String,
+    /// Toggle that determines if the document requires rerendering
+    #[cfg(feature = "syntax_highlighting")]
+    pub needs_rerender: bool,
 }
 
 impl Document {
@@ -76,6 +84,10 @@ impl Document {
             offset: Loc::default(),
             size: size.into(),
             char_ptr: 0,
+            #[cfg(feature = "syntax_highlighting")]
+            render: st!(""),
+            #[cfg(feature = "syntax_highlighting")]
+            needs_rerender: true,
         }
     }
 
@@ -146,10 +158,11 @@ impl Document {
     /// Will return `Err` if the event tried to modifiy data outside the scope of the
     /// document.
     pub fn execute(&mut self, event: Event) -> Result<Status> {
+        let tab_width = self.info.tab_width;
         match event {
             Event::Insert(loc, ch) => {
                 self.goto(loc)?;
-                self.row_mut(loc.y)?.insert(loc.x, ch)?;
+                self.row_mut(loc.y)?.insert(loc.x, ch, tab_width)?;
                 self.modified = true;
                 self.move_right()
             }
@@ -164,7 +177,7 @@ impl Document {
                 Ok(Status::None)
             }
             Event::InsertRow(loc, st) => {
-                self.rows.insert(loc, Row::new(st).link(&mut self.info));
+                self.rows.insert(loc, Row::new(st, self.info.tab_width));
                 self.modified = true;
                 self.goto_y(loc)?;
                 Ok(Status::None)
@@ -372,6 +385,53 @@ impl Document {
         format!("{}{}", " ".repeat(total - num.len()), num)
     }
 
+    /// A helper function that returns info about the document
+    /// in a [`HashMap`] type.
+    ///
+    /// It will return (with keys for the hashmap):
+    /// - Row number: `row`
+    /// - Column number: `column`
+    /// - Total rows: `total`
+    /// - File name (no path): `file`
+    /// - File name (full path): `full_file`
+    /// - File type: `type`
+    /// - Modifed indicator: `modified`
+    /// - File extension: `extension`
+    #[must_use]
+    pub fn status_line_info(&self) -> HashMap<&str, String> {
+        let row = self.loc().y + 1;
+        let total = self.rows.len();
+        let column = self.loc().x;
+        let modified = if self.modified { "[+]" } else { "" };
+        let (full_file, file, ext) = if let Some(name) = &self.info.file {
+            let f = Path::new(&name)
+                .file_name()
+                .unwrap_or_default()
+                .to_str()
+                .unwrap_or(name)
+                .to_string();
+            let e = Path::new(&name)
+                .extension()
+                .unwrap_or_default()
+                .to_str()
+                .unwrap_or("")
+                .to_string();
+            (name.clone(), f, e)
+        } else {
+            (st!("[No Name]"), st!("[No Name]"), st!(""))
+        };
+        let mut info = HashMap::new();
+        info.insert("row", st!(row));
+        info.insert("column", st!(column));
+        info.insert("total", st!(total));
+        info.insert("file", st!(file));
+        info.insert("full_file", st!(full_file));
+        info.insert("type", filetype(&ext).unwrap_or_else(|| st!("Unknown")));
+        info.insert("modified", st!(modified));
+        info.insert("extension", ext);
+        info
+    }
+
     /// Render the document into the correct form
     #[must_use]
     pub fn render(&self) -> String {
@@ -384,6 +444,18 @@ impl Document {
             + line_ending
     }
 
+    /// Render the document into the correct form
+    #[must_use]
+    pub fn render_full(&self) -> String {
+        let line_ending = if self.info.is_dos { "\r\n" } else { "\n" };
+        let mut result = st!("");
+        for r in &self.rows {
+            result.push_str(&r.render_full(self.info.tab_width));
+            result.push_str(line_ending);
+        }
+        result
+    }
+
     /// Shift the cursor back to the nearest grapheme boundary
     fn snap_grapheme(&mut self) -> Result<()> {
         // Collect information
@@ -408,11 +480,12 @@ impl Document {
     }
 
     /// Take raw text and convert it into Row structs
-    fn raw_to_rows(&mut self, text: &str) -> Vec<Row> {
+    #[must_use]
+    pub fn raw_to_rows(&self, text: &str) -> Vec<Row> {
         let text = regex!("(\\r\\n|\\n)$").replace(text, "").to_string();
         let rows: Vec<&str> = regex!("(\\r\\n|\\n)").split(&text).collect();
         rows.iter()
-            .map(|s| Row::new(*s).link(&mut self.info))
+            .map(|s| Row::new(*s, self.info.tab_width))
             .collect()
     }
 
